﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Data;

namespace HeuristicLab.DataPreprocessing {
  public class ManipulationLogic : IManipulationLogic {
    private ITransactionalPreprocessingData preprocessingData;
    private IStatisticsLogic statisticInfo;
    private ISearchLogic searchLogic;

    public ManipulationLogic(ITransactionalPreprocessingData _prepocessingData, ISearchLogic theSearchLogic, IStatisticsLogic theStatisticsLogic) {
      preprocessingData = _prepocessingData;
      searchLogic = theSearchLogic;
      statisticInfo = theStatisticsLogic;
    }

    public void ReplaceIndicesByValue<T>(int columnIndex, IEnumerable<int> rowIndices, T value) {
      foreach (int index in rowIndices) {
        preprocessingData.SetCell<T>(columnIndex, index, value);
      }
    }

    public void ReplaceIndicesByAverageValue(Dictionary<int, List<int>> cells) {
      preprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          double average = statisticInfo.GetAverage(column.Key);
          ReplaceIndicesByValue<double>(column.Key, column.Value, average);
        }
      });
    }

    public void ReplaceIndicesByMedianValue(Dictionary<int, List<int>> cells) {
      preprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          double median = statisticInfo.GetMedian(column.Key);
          ReplaceIndicesByValue<double>(column.Key, column.Value, median);
        }
      });
    }

    public void ReplaceIndicesByRandomValue(Dictionary<int, List<int>> cells) {
      preprocessingData.InTransaction(() => {
        Random r = new Random();

        foreach (var column in cells) {
          double max = statisticInfo.GetMax<double>(column.Key);
          double min = statisticInfo.GetMin<double>(column.Key);
          double randMultiplier = (max - min);
          foreach (int index in column.Value) {
            double rand = r.NextDouble() * randMultiplier + min;
            preprocessingData.SetCell<double>(column.Key, index, rand);
          }
        }
      });
    }

    public void ReplaceIndicesByLinearInterpolationOfNeighbours(Dictionary<int, List<int>> cells) {
      preprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          int countValues = preprocessingData.GetValues<double>(column.Key).Count();
          foreach (int index in column.Value) {
            // dont replace first or last values
            if (index > 0 && index < countValues) {
              int prevIndex = indexOfPrevPresentValue(column.Key, index);
              int nextIndex = indexOfNextPresentValue(column.Key, index);

              // no neighbours found
              if (prevIndex < 0 && nextIndex >= countValues) {
                continue;
              }
              double prev = preprocessingData.GetCell<double>(column.Key, prevIndex);
              double next = preprocessingData.GetCell<double>(column.Key, nextIndex);

              int valuesToInterpolate = nextIndex - prevIndex;

              double interpolationStep = (prev + next) / valuesToInterpolate;

              for (int i = prevIndex; i < nextIndex; ++i) {
                double interpolated = prev + (interpolationStep * (i - prevIndex));
                preprocessingData.SetCell<double>(column.Key, i, interpolated);
              }
            }
          }
        }
      });
    }

    private int indexOfPrevPresentValue(int columnIndex, int start) {
      int offset = start - 1;
      while (offset >= 0 && searchLogic.IsMissingValue(columnIndex, offset)) {
        offset--;
      }

      return offset;
    }

    private int indexOfNextPresentValue(int columnIndex, int start) {
      int offset = start + 1;
      while (offset < preprocessingData.Rows && searchLogic.IsMissingValue(columnIndex, offset)) {
        offset++;
      }

      return offset;
    }

    public void ReplaceIndicesByMostCommonValue(Dictionary<int, List<int>> cells) {
      preprocessingData.InTransaction(() => {
        foreach (var column in cells) {
          if (preprocessingData.IsType<double>(column.Key)) {
            ReplaceIndicesByValue<double>(column.Key, column.Value, statisticInfo.GetMostCommonValue<double>(column.Key));
          } else if (preprocessingData.IsType<string>(column.Key)) {
            ReplaceIndicesByValue<string>(column.Key, column.Value, statisticInfo.GetMostCommonValue<string>(column.Key));
          } else if (preprocessingData.IsType<DateTime>(column.Key)) {
            ReplaceIndicesByValue<DateTime>(column.Key, column.Value, statisticInfo.GetMostCommonValue<DateTime>(column.Key));
          } else {
            throw new ArgumentException("column with index: " + column.Key + " contains a non supported type.");
          }
        }
      });
    }

    public void ShuffleWithRanges(IEnumerable<IntRange> ranges) {
      // init random outside loop
      Random random = new Random();

      preprocessingData.InTransaction(() => {
        // process all given ranges - e.g. TrainingPartition, Trainingpartition
        foreach (IntRange range in ranges) {
          List<Tuple<int, int>> shuffledIndices = new List<Tuple<int, int>>();

          // generate random indices used for shuffeling each column
          for (int i = range.End; i > range.Start; --i) {
            int rand = random.Next(range.Start, i);
            shuffledIndices.Add(new Tuple<int, int>(i, rand));
          }

          ReOrderToIndices(shuffledIndices);
        }
      });
    }

    public void ReOrderToIndices(IEnumerable<int> indices) {
      List<Tuple<int, int>> indicesTuple = new List<Tuple<int, int>>();

      for (int i = 0; i < indices.Count(); ++i) {
        indicesTuple.Add(new Tuple<int, int>(i, indices.ElementAt(i)));
      }

      ReOrderToIndices(indicesTuple);
    }

    public void ReOrderToIndices(IList<System.Tuple<int, int>> indices) {
      preprocessingData.InTransaction(() => {
        for (int i = 0; i < preprocessingData.Columns; ++i) {
          if (preprocessingData.IsType<double>(i)) {
            reOrderToIndices<double>(i, indices);
          } else if (preprocessingData.IsType<string>(i)) {
            reOrderToIndices<string>(i, indices);
          } else if (preprocessingData.IsType<DateTime>(i)) {
            reOrderToIndices<DateTime>(i, indices);
          }
        }
      });
    }

    private void reOrderToIndices<T>(int columnIndex, IList<Tuple<int, int>> indices) {

      List<T> originalData = new List<T>(preprocessingData.GetValues<T>(columnIndex));

      // process all columns equally
      foreach (Tuple<int, int> index in indices) {
        int originalIndex = index.Item1;
        int replaceIndex = index.Item2;

        T replaceValue = originalData.ElementAt<T>(replaceIndex);
        preprocessingData.SetCell<T>(columnIndex, originalIndex, replaceValue);
      }
    }
  }
}
